# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Root program for metapack programs

"""

import argparse
import logging
import sys
from pkg_resources import (
    DistributionNotFound,
    get_distribution,
    iter_entry_points
)

from metapack import Downloader
from metapack.cli.core import cli_init, warn

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass


def base_parser():
    """Entry program for running Metapack commands.
    """
    parser = argparse.ArgumentParser(
        prog='mp',
        description=base_parser.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument('--exceptions', '-e', default=False, action='store_true',
                        help='Show full stack tract for some unhandled exceptions')

    parser.add_argument('--debug', '-d', default=False, action='store_true',
                        help='Turn on debug logging ( Basic Config ) ')

    parser.add_argument('--quiet', '-q', default=False, action='store_true',
                        help='Display less output')

    parser.add_argument('--no-cache', '-n', default=False, action='store_true',
                        help='Ignore the download cache')

    subparsers = parser.add_subparsers(help='Commands')

    for ep in iter_entry_points(group='mt.subcommands'):
        f = ep.load()
        f(subparsers)

    return parser


def setup_downloader(args):
    if args.no_cache:
        downloader = Downloader.get_instance()
        downloader.use_cache = False


def mp(args=None, do_cli_init=True):
    from .core import err

    try:
        parser = base_parser()

        parsed_args = parser.parse_args(args)

        if do_cli_init:
            cli_init(log_level=logging.DEBUG if parsed_args.debug else
                     logging.WARNING if parsed_args.quiet else
                     logging.INFO)

        setup_downloader(parsed_args)
    except KeyboardInterrupt:
        warn('Keyboard Interrupt')
        return 0

    try:
        parsed_args.run_command  # Happens when no commands are specified
    except AttributeError:
        parser.print_help()
        sys.exit(2)

    try:
        return parsed_args.run_command(parsed_args)
    except KeyboardInterrupt as e:
        if parsed_args.exceptions:
            raise e
        warn('Keyboard Interrupt')

    except Exception as e:
        if parsed_args.exceptions:
            raise e
        else:
            if e.__cause__:
                err(f"{e.__class__.__name__} {str(e)}\nCaused by: {type(e.__cause__)} {str(e.__cause__)}")
            else:
                err(f"{e.__class__.__name__}  {str(e)}")



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""
import argparse
import sys
from pkg_resources import (
    DistributionNotFound,
    get_distribution,
    iter_entry_points
)

from tabulate import tabulate

from metapack.cli.core import err, prt
from metapack.package import Downloader
from metapack.util import iso8601_duration_as_seconds

from .core import MetapackCliMemo as _MetapackCliMemo

downloader = Downloader.get_instance()


class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)
        self


def info_args(subparsers):
    """
    Entry program for running Metapack commands.
    """

    parser = subparsers.add_parser(
        'info',
        help='Print info about a package ',
        description=info_args.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.set_defaults(run_command=info)

    group = parser.add_mutually_exclusive_group()

    group.add_argument('-n', '--name', default=False, action='store_true',
                       help="Print the name, with version")
    group.add_argument('-N', '--root-name', default=False, action='store_true',
                       help="Print the name, without the version")

    group.add_argument('-r', '--resources', default=False, action='store_true',
                       help="List the resources in the package")

    group.add_argument('-s', '--schema', default=False, action='store_true',
                       help="Print a table of the common schema for all resources, or "
                            "if the metatab file ref has a resource, only that one")

    group.add_argument('-R', '--row-table', default=False, action='store_true',
                       help="Print the row-processor table, including transforms and valuetypes")

    group.add_argument('-p', '--package-url', default=False, action='store_true',
                       help="Print the package url")

    group.add_argument('-P', '--package-root', default=False, action='store_true',
                       help="Print the package root url")

    group.add_argument('-T', '--value-types', default=False, action='store_true',
                       help='Print a list of available value types')

    group.add_argument('-t', '--transforms', default=False, action='store_true',
                       help='Print a list of available transform functions')

    group.add_argument('-U', '--next-update', default=False, action='store_true',
                       help='Print the date of the next update, based on Root.Modified and Root.UpdateFrequency')

    parser.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")


def info(args):
    from metapack.exc import MetatabFileNotFound

    m = MetapackCliMemo(args, downloader)

    rc = None

    try:
        if m.args.name:
            prt(m.doc.name)

        elif m.args.resources:
            list_rr(m.doc)

        elif m.args.root_name:
            prt(m.doc.as_version(None))

        elif m.args.package_url:
            prt(m.package_url)

        elif m.args.package_root:
            prt(m.package_root)

        elif m.args.schema:
            dump_schemas(m)

        elif m.args.row_table:
            dump_rptable(m)

        elif m.resource:
            resource_info(m)

        elif args.value_types:
            print_value_types(m)

        elif args.transforms:
            print_transforms(m)

        elif args.next_update:
            rc = next_update(m)

        else:
            prt(m.doc.name)

    except MetatabFileNotFound:
        err('No metatab file found')

    return rc


def resource_info(m):
    from rowgenerators.exceptions import RowGeneratorError

    r = m.get_resource()

    ru = r.resolved_url
    rsu = ru.get_resource()
    tu = rsu.get_target()
    rows = [
        ('Value', r.url),
        ("Expanded", r.expanded_url),
        ('Resolved', ru),
        ('URL Resource', rsu),
        ('URL path', tu),
        ('Target Path', tu.fspath)
    ]

    props = ['fragment_query', 'resource_file', 'resource_format',
             'target_file', 'segment', 'target_format',
             'encoding', 'headers', 'start', 'end']

    rows += [(p, getattr(tu, p, '')) for p in props]

    rows.append(('URL Type', type(ru)))

    try:
        rows.append(('Generator', type(ru.generator)))
    except RowGeneratorError:
        rows.append(('Generator', "Error: Can't locate"))

    prt(tabulate(rows))


def print_versions(m):
    main_packages = ('metapack', 'metatab', 'metatabdecl', 'rowgenerators', 'publicdata', 'tableintuit')

    packages = []
    for pkg_name in main_packages:
        try:
            d = get_distribution(pkg_name)
            packages.append([d.project_name, d.version])

        except (DistributionNotFound, ModuleNotFoundError):
            # package is not installed

            pass

    prt(tabulate(packages, headers='Package Version'.split()))
    prt('')
    prt(tabulate([(ep.name, ep.dist) for ep in iter_entry_points(group='mt.subcommands')],
                 headers='Subcommand Package Version'.split()))


def list_rr(doc):
    d = []
    for r in doc.resources():
        d.append(('Resource', r.name, r.url))

    if d:
        prt('== Resources ==')
        prt(tabulate(d, 'Type Name Url'.split()))
        prt('')

    d = []
    for r in doc.references():
        d.append(('Reference', r.name, r.url))

    if d:
        prt('== References ==')
        prt(tabulate(d, 'Type Name Url'.split()))


def get_resource(m):
    r = m.doc.resource(m.resource)

    if not r:
        prt('\nSelect a resource to display the schema for:\n')
        list_rr(m.doc)
        prt('')
        prt("Specify the resource as a fragment, escaping it if the '#' is the first character. For instance: ")
        prt("  mp info -s .#resource_name")
        prt('')
        sys.exit(0)

    return r


def dump_schemas(m):
    r = get_resource(m)

    st = r.schema_term

    if not st:
        err("No schema term for resource {}".format(r.name))

    rows_about_columns = []

    has_vt = any(c.get_value('valuetype') for c in st.find('Table.Column'))

    if has_vt:
        headers = 'Name AltName DataType ValueType Description'.split()

        def cols(c):
            return (c.name, c.get_value('altname'), c.get_value('datatype'), c.get_value('valuetype'),
                    c.get_value('description'))
    else:
        headers = 'Name AltName DataType Description'.split()

        def cols(c):
            return (c.name, c.get_value('altname'), c.get_value('datatype'), c.get_value('description'))

    for c in st.find('Table.Column'):
        rows_about_columns.append(cols(c))

    prt(tabulate(rows_about_columns, headers=headers))


def dump_rptable(m):
    r = get_resource(m)

    print(r.row_processor_table())


def print_declare(m):
    from metatab.util import declaration_path

    prt(declaration_path('metatab-latest'))


def print_value_types(m):
    from rowgenerators.valuetype import value_types

    rows = [(k, v.__name__, v.__doc__) for k, v in value_types.items()]

    print(tabulate(sorted(rows), headers='Code Class Description'.split()))


def print_transforms(m):
    env = m.doc.env

    print(env)
    return

    # rows = [(k, v.__name__, v.__doc__) for k, v in value_types.items()]

    # print(tabulate(sorted(rows), headers='Code Class Description'.split()))


def next_update(m):
    from datetime import datetime, timedelta
    from dateutil.parser import parse

    last_mod = m.doc['Root'].find_first_value('Root.Modified')

    if not last_mod:
        prt('No Root.Modified')
        return 2

    last_mod = parse(last_mod)

    uf = m.doc['Root'].find_first_value('Root.UpdateFrequency')

    now = datetime.now().date()

    if not uf:
        prt(now.isoformat())
        return 1

    td = timedelta(seconds=iso8601_duration_as_seconds(uf))

    update_time = (last_mod + td).date()

    if update_time <= now:
        print('UPDATE', update_time.isoformat())
        return 0
    else:
        print('OK', update_time.isoformat())
        return 1



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Metapack CLI program for creating new metapack package directories
"""

import argparse
import sys

from metapack import Downloader
from metapack.cli.core import MetapackCliMemo, list_rr, prt

from .core import MetapackCliMemo as _MetapackCliMemo

downloader = Downloader.get_instance()

class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)

def open_args(subparsers):
    """
    The `mp open` command will open a resource with the system application, such as Excel or OpenOffice

    """
    parser = subparsers.add_parser(
        'open',
        help='open a CSV resource with a system application',
        description=open_args.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.set_defaults(run_command=open_cmd)

    parser.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

    return parser


def open_cmd(args):

    m = MetapackCliMemo(args, downloader)

    r = m.get_resource()

    doc = m.doc

    if not r:
        list_rr(doc)
        sys.exit(1)



""" Example pylib functions"""


def row_generator(resource, doc, env, *args, **kwargs):
    """ An example row generator function.

    Reference this function in a Metatab file as the value of a Datafile:

            Datafile: python:pylib#row_generator

    The function must yield rows, with the first being headers, and subsequenct rows being data.

    :param resource: The Datafile term being processed
    :param doc: The Metatab document that contains the term being processed
    :param args: Positional arguments passed to the generator
    :param kwargs: Keyword arguments passed to the generator
    :return:


    The env argument is a dict with these environmental keys:

    * CACHE_DIR
    * RESOURCE_NAME
    * RESOLVED_URL
    * WORKING_DIR
    * METATAB_DOC
    * METATAB_WORKING_DIR
    * METATAB_PACKAGE

    It also contains key/value pairs for all of the properties of the resource.

    """

    yield 'a b c'.split()

    for i in range(10):
        yield [i, i*2, i*3]


def example_transform(v, row, row_n, i_s, i_d, header_s, header_d,scratch, errors, accumulator):
    """ An example column transform.

    This is an example of a column transform with all of the arguments listed. An real transform
    can omit any ( or all ) of these, and can supply them in any order; the calling code will inspect the
    signature.

    When the function is listed as a transform for a column, it is called for every row of data.

    :param v: The current value of the column
    :param row: A RowProxy object for the whiole row.
    :param row_n: The current row number.
    :param i_s: The numeric index of the source column
    :param i_d: The numeric index for the destination column
    :param header_s: The name of the source column
    :param header_d: The name of the destination column
    :param scratch: A dict that can be used for storing any values. Persists between rows.
    :param errors: A dict used to store error messages. Persists for all columns in a row, but not between rows.
    :param accumulator: A dict for use in accumulating values, such as computing aggregates.
    :return: The final value to be supplied for the column.
    """

    return str(v)+'-foo'



# Copyright (c) 2019 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

""" """

from rowgenerators import Downloader as _Downloader
from rowgenerators import parse_app_url

from metapack.appurl import MetapackUrl

DEFAULT_CACHE_NAME = 'metapack'


class Downloader(_Downloader):
    """"Local version of the downloader. Also should be used as the source of the cache"""

    ok = True

    def __init__(self, cache=None, account_accessor=None, logger=None, working_dir='', callback=None):
        from rowgenerators import get_cache
        super().__init__(cache or get_cache('metapack'),
                         account_accessor, logger, working_dir, callback)

    def download(self, url):
        return super().download(url)


def open_package(ref, downloader=None):
    from metapack.doc import MetapackDoc

    if downloader is None:
        downloader = Downloader()

    if isinstance(ref, MetapackUrl):
        p = MetapackDoc(ref, downloader=downloader)
        p.default_resource = None
        return p

    else:

        ref = str(ref)

        if ref.startswith('index:'):
            from metapack.appurl import SearchUrl
            SearchUrl.initialize()
            ref = str(parse_app_url(ref).resolve())

        u = MetapackUrl(ref, downloader=downloader)

        p = MetapackDoc(u, downloader=downloader)
        p.default_resource = u.resource_name
        return p



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

"""

from os.path import basename, dirname, join

from metatab import DEFAULT_METATAB_FILE, LINES_METATAB_FILE
from rowgenerators import Url, parse_app_url
from rowgenerators.appurl.file.file import FileUrl
from rowgenerators.appurl.util import file_ext
from rowgenerators.appurl.web.web import WebUrl
from rowgenerators.exceptions import AppUrlError, DownloadError

from metapack.exc import MetapackError, ResourceError

METATAB_FILES = (DEFAULT_METATAB_FILE, LINES_METATAB_FILE, DEFAULT_METATAB_FILE.replace('.csv', '.ipynb'))

SIMPLE_FILE_FORMATS = ('csv', 'txt', 'ipynb')


class _MetapackUrl(object):

    def exists(self):

        if self.inner.proto == 'file':
            return self.inner.exists()
        else:
            # Hack, really ought to implement exists() for web urls.
            return True

    @property
    def resource_name(self):
        return self._parts['target_file']  # Must be underlying property

    def absolute(self):
        return self.inner.absolute()


def is_metapack_url(u):
    return isinstance(u, _MetapackUrl)


class MetapackDocumentUrl(Url, _MetapackUrl):

    def __init__(self, url=None, downloader=None, **kwargs):

        assert downloader

        super().__init__(url, downloader=downloader, **kwargs)

        self.scheme_extension = 'metapack'

        self._set_target_file()

    @classmethod
    def _match(cls, url, **kwargs):
        raise MetapackError("This class should not be constructed through matching")

    def _set_target_file(self):

        if self.resource_file in METATAB_FILES:
            return  # self.resource_file

        elif self.resource_format in SIMPLE_FILE_FORMATS:
            return  # self.resource_file

        elif self.resource_format == 'xlsx':
            self.target_file = 'meta'
            return

        elif self.resource_format == 'zip':
            self.target_file = DEFAULT_METATAB_FILE
            return

        elif file_ext(self.resource_file) not in SIMPLE_FILE_FORMATS:
            self.path = join(self.path, DEFAULT_METATAB_FILE)

        else:
            return self.resource_file

    @property
    def doc(self):
        """Return the metatab document for the URL"""
        from metapack import MetapackDoc
        t = self.get_resource().get_target()
        return MetapackDoc(t, package_url=self.package_url)

    @property
    def generator(self):

        from rowgenerators import get_generator

        # HACK! This used to be:
        # t = self.get_resource().get_target().inner
        # I don't know why, but stripping off the proto is problematic
        # for other cases.

        t = self.get_resource().get_target()

        return get_generator(t)

    def resolve_url(self, resource_name):
        """Return a URL to a local copy of a resource, suitable for get_generator()"""

        if self.target_format == 'csv' and self.target_file != DEFAULT_METATAB_FILE:
            # For CSV packages, need to get the package and open it to get the resoruce URL, becuase
            # they are always absolute web URLs and may not be related to the location of the metadata.
            s = self.get_resource()
            rs = s.doc.resource(resource_name)
            return parse_app_url(rs.url)
        else:
            jt = self.join_target(resource_name)
            rs = jt.get_resource()
            t = rs.get_target()
        return t

    @property
    def metadata_url(self):

        if not basename(self.resource_url):
            return self.clone(path=join(self.path, DEFAULT_METATAB_FILE))
        else:
            return self.clone()

    @property
    def package_url(self):
        """Return the package URL associated with this metadata"""

        if self.resource_file == DEFAULT_METATAB_FILE or self.target_format in ('txt', 'ipynb'):
            u = self.inner.clone().clear_fragment()
            u.path = dirname(self.path) + '/'

        else:
            u = self.clear_fragment()

        return MetapackPackageUrl(str(u), downloader=self._downloader)

    def get_resource(self):

        if self.scheme == 'file':
            u = self
        else:
            u = WebUrl(str(self), downloader=self._downloader).get_resource()

        return MetapackDocumentUrl(str(u), downloader=self._downloader)

    def get_target(self):
        u = self.inner.get_target().clone(scheme_extension=self.proto)
        u.scheme_extension = self.scheme_extension
        return parse_app_url(str(u), downloader=self._downloader)

    def join_target(self, tf):

        if self.target_file in METATAB_FILES:
            return self.inner.join_dir(tf)
        else:
            return self.inner.join_target(tf)

    @property
    def resource(self):
        """Return the Metapack resource, different from the URL resource returned by get_resource"""

        r = self.doc.resource(self.resource_name)
        return r


class MetapackPackageUrl(FileUrl, _MetapackUrl):
    """Special version of MetapackUrl for package urls, which never have a fragment"""

    def __init__(self, url=None, downloader=None, **kwargs):

        kwargs['scheme_extension'] = 'metapack'

        super().__init__(url, downloader=downloader, **kwargs)

        assert self._downloader

    @property
    def package_url(self):
        return self

    @property
    def metadata_url(self):

        return MetapackDocumentUrl(str(self.clone().clear_fragment()), downloader=self._downloader).metadata_url

    @property
    def doc(self):
        """Return the metatab document for the URL"""
        return self.metadata_url.doc

    def join_resource_name(self, v):
        """Return a MetapackResourceUrl that includes a reference to the resource. Returns a
        MetapackResourceUrl, which will have a fragment """
        d = self.dict
        self.target_file = v
        return MetapackResourceUrl(downloader=self._downloader, **d)

    def join_resource_path(self, v):
        """Return a regular AppUrl that combines the package with a URL path """
        return self.inner.join(v)

    def join_target(self, tf):
        """Like join(), but returns the inner URL, not a package url class"""
        if self.target_file in METATAB_FILES:
            return self.inner.join_dir(tf)
        else:
            return self.inner.join_target(tf)

    def resolve_url(self, resource_name):
        """Return a URL to a local copy of a resource, suitable for get_generator()

        For Package URLS, resolution involves generating a URL to a data file from the package URL and the
        value of a resource. The resource value, the url, can be one of:

        - An absolute URL, with a web scheme
        - A relative URL, relative to the package, with a file scheme.

        URLs with non-file schemes are returned. File scheme are assumed to be relative to the package,
        and are resolved according to the type of resource.

        """

        u = parse_app_url(resource_name)

        if u.scheme != 'file':
            t = u

        elif self.target_format == 'csv' and self.target_file != DEFAULT_METATAB_FILE:
            # Thre are two forms for CSV package URLS:
            # - A CSV package, which can only have absolute URLs
            # - A Filesystem package, which can have relative URLs.

            # The complication is that the filsystem package usually has a metadata file named
            # DEFAULT_METATAB_FILE, which can distinguish it from a CSV package, but it's also possible
            # to have a filesystem package with a non standard package name.

            # So, this clause can happed for two cases: A CSV package or a Filesystem package with a nonstandard
            # metadata file name.

            # For CSV packages, need to get the package and open it to get the resource URL, because
            # they are always absolute web URLs and may not be related to the location of the metadata.

            s = self.get_resource()
            rs = s.metadata_url.doc.resource(resource_name)
            if rs is not None:
                t = parse_app_url(rs.url)
            else:
                raise ResourceError("No resource for '{}' in '{}' ".format(resource_name, self))

        else:
            jt = self.join_target(resource_name)

            try:
                rs = jt.get_resource()
            except DownloadError:
                raise DownloadError(
                    "Failed to download resource for '{}' for '{}' in '{}'".format(jt, resource_name, self))
            t = rs.get_target()

        return t

    @property
    def inner(self):
        return super().inner


class MetapackResourceUrl(FileUrl, _MetapackUrl):
    def __init__(self, url=None, downloader=None, base_url=None, **kwargs):

        kwargs['proto'] = 'metapack'

        super().__init__(url, downloader=downloader, **kwargs)
        self.scheme_extension = 'metapack'

        # Expand the path in the same was as the document URL

        # Using .clone() causes recursion
        d = self.dict
        # Remove the resource
        d['target_file'] = None
        d['target_segment'] = None

        md = MetapackDocumentUrl(None, downloader=downloader, **d)
        self.path = md.path

        self.base_url = base_url

    @classmethod
    def _match(cls, url, **kwargs):
        raise MetapackError("This class should not be contructed through matching")

    @property
    def doc(self):
        """Return the metatab document for the URL"""
        return self.metadata_url.doc

    @property
    def metadata_url(self):
        assert self._downloader
        return MetapackDocumentUrl(str(self.clone().clear_fragment()), downloader=self._downloader).metadata_url

    @property
    def package_url(self):
        """Return the package URL associated with this metadata"""
        return MetapackDocumentUrl(str(self.clear_fragment()), downloader=self._downloader).package_url

    def get_resource(self):
        if self.scheme == 'file':
            return self
        else:
            u = WebUrl(str(self), downloader=self._downloader)
            r = u.get_resource()
            mru = MetapackResourceUrl(str(r), base_url=self, downloader=self._downloader)
            return mru

    def get_target(self):
        return self.resource.resolved_url.get_resource().get_target()

    @property
    def generator(self):
        # Use fragment b/c it could be target_file, for .zip, or target_segment, for .xlsx
        return self.resource

    @property
    def resource(self):

        r = self.doc.resource(self.resource_name)

        if not r:
            r = self.doc.reference(self.resource_name)

        return r


# Would have made this a function, but it needs to be a class to have the match() method
class MetapackUrl(Url):
    """Implements __new__ to return either a  MetapackResourceUrl or a MetapackDocumentUrl"""

    match_priority = WebUrl.match_priority - 1

    def __new__(cls, url=None, downloader=None, **kwargs):

        assert downloader

        u = Url(url, **kwargs)

        if u._parts['target_file']:  # must be underlying property, not .target_file
            return MetapackResourceUrl(url, downloader, **kwargs)
        else:
            return MetapackDocumentUrl(url, downloader, **kwargs)

    @classmethod
    def _match(cls, url, **kwargs):
        """Return True if this handler can handle the input URL"""
        return url.proto in ('metapack', 'metatab')


class SearchUrl(Url):
    """Look up a url using a list of callbacks. """

    match_priority = FileUrl.match_priority - 10

    search_callbacks = []

    _search_initialized = False

    def __init__(self, url=None, **kwargs):
        kwargs['scheme'] = 'index'
        super().__init__(url, **kwargs)

    @classmethod
    def _match(cls, url, **kwargs):
        return url.scheme == 'search' or url.scheme == 'index'

    @classmethod
    def register_search(cls, f):
        """Register a function that will lookup the search URL and return a URL if found"""

        cls.search_callbacks.append(f)

    @classmethod
    def initialize(cls):
        if SearchUrl._search_initialized is False:
            from metapack.package import Downloader  # Breaks inclusion cycle
            try:
                search_func = cls.search_json_indexed_directory(Downloader.get_instance().cache.getsyspath('/'))
                SearchUrl.register_search(search_func)
            except AppUrlError:
                pass

            SearchUrl._search_initialized = True

    @staticmethod
    def search_json_indexed_directory(directory):
        """Return a search function for searching a directory of packages, which has an index.json file
        created by the `mp install file` command.

        This will only search the issued index; it will not return results for the source index
        """

        from metapack.index import SearchIndex, search_index_file

        idx = SearchIndex(search_index_file())

        def _search_function(url):

            packages = idx.search(url, format='issued')

            if not packages:
                return None

            package = packages.pop(0)

            try:
                resource_str = '#' + url.resource_name if url.resource_name else ''

                return parse_app_url(package['url'] + resource_str, downloader=url.downloader)
            except KeyError:
                return None

        return _search_function

    def search(self):
        """Search for a url by returning the value from the first callback that
        returns a non-None value"""

        for cb in SearchUrl.search_callbacks:

            try:
                v = cb(self)
                if v is not None:
                    return v
            except Exception:
                raise

    def resolve(self):
        u = self.search()

        if not u:
            raise AppUrlError("Search URL failed to resolve reference to {} ".format(str(self)))

        # u.set_fragment(self.fragment)

        return u

    def get_resource(self):

        return self.resolve().get_resource()

    def get_target(self):
        return self

    @property
    def resource_name(self):
        """Name of a Metatab resource ( not the URL resource ) which is specified in the fragment"""

        return self._parts.get('target_file')



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""

from pkg_resources import (
    DistributionNotFound,
    get_distribution,
    iter_entry_points
)

from tabulate import tabulate

from metapack.cli.core import err, prt
from metapack.package import *

downloader = Downloader.get_instance()


def config_args(subparsers):
    parser = subparsers.add_parser(
        'config',
        help='Print configuration information about the metapack installation'
    )

    parser.set_defaults(run_command=config)

    parser.add_argument('-v', '--version', default=False, action='store_true',
                        help='Print Metapack versions')

    parser.add_argument('-V', '--versions', default=False, action='store_true',
                        help='Print version of several important packages')

    parser.add_argument('-c', '--cache', default=False, action='store_true',
                        help='Print the location of the cache')

    parser.add_argument('-d', '--declare', default=False, action='store_true',
                        help='Print the location of the default declarition document')

    parser.add_argument('-m', '--materialized', default=False, action='store_true',
                        help='Print the location of the materialized data cache')

    parser.add_argument('-t', '--value-types', default=False, action='store_true',
                        help='Print a list of available value types')


def config(args):
    from metapack.exc import MetatabFileNotFound

    try:

        if args.version:
            prt(get_distribution('metapack'))

        elif args.cache:
            from shlex import quote
            prt(downloader.cache.getsyspath('/'))

        elif args.materialized:
            from metapack.util import get_materialized_data_cache
            prt(get_materialized_data_cache())

        elif args.versions:
            print_versions()

        elif args.declare:
            print_declare()

        elif args.value_types:
            print_value_types()

        else:
            print_versions()

    except MetatabFileNotFound:
        err('No metatab file found')


def print_versions():
    from tabulate import tabulate

    main_packages = ('metapack', 'metatab', 'metatabdecl', 'rowgenerators', 'tableintuit')

    packages = []
    for pkg_name in main_packages:
        try:
            d = get_distribution(pkg_name)
            packages.append([d.project_name, d.version])

        except (DistributionNotFound, ModuleNotFoundError) as e:
            # package is not installed
            print(e)

    prt(tabulate(packages, headers='Package Version'.split()))
    prt('')
    prt(tabulate([(ep.name, ep.dist) for ep in iter_entry_points(group='mt.subcommands')],
                 headers='Subcommand Package Version'.split()))


def list_rr(doc):
    d = []
    for r in doc.resources():
        d.append(('Resource', r.name, r.url))

    if d:
        prt('== Resources ==')
        prt(tabulate(d, 'Type Name Url'.split()))
        prt('')

    d = []
    for r in doc.references():
        d.append(('Reference', r.name, r.url))

    if d:
        prt('== References ==')
        prt(tabulate(d, 'Type Name Url'.split()))


def print_declare():
    from metatab.util import declaration_path

    prt(declaration_path('metatab-latest'))


def print_value_types():
    from rowgenerators.valuetype import value_types

    rows = [(k, v.__name__, v.__doc__) for k, v in value_types.items()]

    print(tabulate(sorted(rows), headers='Code Class Description'.split()))



import unittest

from support import open_package, test_data
from tabulate import tabulate


def ds_hash(r):
    import hashlib

    m = hashlib.md5()

    for row in r:
        for col in row:
            m.update(str(col).encode('utf8'))

    return m.hexdigest()

class TestPackages(unittest.TestCase):
    """"""

    def setUp(self):
        import warnings
        warnings.simplefilter('ignore')

    def test_iterators(self):
        from itertools import chain
        pkg = open_package('example.com-iterators')

        print('== References')

        for r in chain(pkg.references(), pkg.resources()):

            if r.get_value('Iterator'):
                itr = getattr(r,r.get_value('Iterator') )
            else:
                itr = iter(r)


            data = list(itr)
            hash = ds_hash(data)

            if hash != r.get_value('Hash'):
                print(tabulate(data))
                print(r.name, hash)



if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Tests for CLI programs
"""


import unittest

from metapack import open_package
from metapack.cli.test.core import delete_contents

from .core import exec_cmd


class TestCli(unittest.TestCase):

    def setUp(self):
        import tempfile
        import os

        super().setUp()

        self.env_dir = os.getenv('METAPACK_TEST_DIR')

        if not self.env_dir:
            self.temp_dir = tempfile.mkdtemp()
        else:
            self.temp_dir = None
            delete_contents(self.env_dir)

        self.test_dir = self.env_dir or self.temp_dir

        os.chdir(self.test_dir)

    def tearDown(self):

        super().tearDown()

        if self.temp_dir:
            delete_contents(self.temp_dir)

    def test_basic_build(self):
        from os.path import exists, join
        from os import chdir

        out, err, exc = exec_cmd('new -o example.com -d dataset')

        self.assertIsNone(exc)

        pkg_dir = join(self.test_dir, 'example.com-dataset')

        self.assertTrue(exists(pkg_dir))

        with open_package(pkg_dir) as pkg:
            pkg.find_first('Root.Version').value = '1.1.1'
            pkg['Resources'].clean()
            pkg['Resources'].new_term('Datafile',
                                      'http://public.source.civicknowledge.com/example.com/sources/renter_cost.csv',
                                      name='renter_cost')

        chdir(pkg_dir)

        out, err, exc = exec_cmd('build -f')

        self.assertIsNotNone(exc) # Should be SystemExit

        out, err, exc = exec_cmd('pack -s')

        self.assertIsNone(exc,err)

        out, err, exc = exec_cmd('build -f -e -c -z')

        self.assertIsNone(exc)

        self.assertTrue(exists(join(pkg_dir,'_packages/example.com-dataset-1.1.1.zip')))


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""

import argparse
import tempfile
from os.path import join

from rowgenerators import parse_app_url
from rowgenerators.appurl.file import FileUrl
from rowgenerators.appurl.web import S3Url
from rowgenerators.exceptions import RowGeneratorError
from tabulate import tabulate

from metapack.constants import PACKAGE_PREFIX
from metapack.exc import MetatabFileNotFound
from metapack.index import SearchIndex, search_index_file
from metapack.package import Downloader, open_package

from .core import err, prt

downloader = Downloader.get_instance()


def index_args(subparsers):
    """Index packages for searching.

    The index file is a JSON file, which is by default index.json in the cache.
    The file can be moved by setting the METAPACK_SEARCH_INDEX environmental variable.

    """
    parser = subparsers.add_parser(
        'index',
        description=index_args.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.set_defaults(run_command=index)

    parser.add_argument('-l', '--list', default=False, action='store_true',
                        help="List the packages that would be indexed")

    parser.add_argument('-c', '--clear', default=False, action='store_true',
                        help="Clear the index")

    parser.add_argument('-C', '--config', default=False, action='store_true',
                        help="Show the location of the index file")

    parser.add_argument('-f', '--file', help="Alternative path to index file")

    parser.add_argument('-w', '--write', help="After creating an index file from S3 and  write back the index file",
                        action='store_true', required=False)

    parser.add_argument('-D', '--dump',
                        help="Dump a Metatab formatted version of the index")

    parser.add_argument('-L', '--load',
                        help="Load in a Metatab formatted index file ")

    parser.add_argument('-p', '--profile', help="Name of a BOTO or AWS credentials profile, for S3 indexing",
                        required=False)

    parser.add_argument('-r', '--result', action='store_true', default=False,
                        help="If mp -q flag set, still report results")

    parser.add_argument('metatab_url', nargs='?', default='./',
                        help='URL to a metatab package or container for packages')


def walk_packages(args, u):
    from os import walk
    from os.path import islink, join, isdir

    seen = set()

    # Try all files as packages
    if not isdir(u.path):
        try:
            yield open_package(u.path)
        except (RowGeneratorError, MetatabFileNotFound):
            pass

        return

    for root, dirs, files in walk(u.path):

        try:
            p = open_package(root)

            # This was a package, so only recurse if it is a source package and has a _packages dir

            if PACKAGE_PREFIX in dirs:
                del dirs[:]
                dirs.append(PACKAGE_PREFIX)
            else:
                del dirs[:]

            if str(p.ref) not in seen:
                yield p

            seen.add(str(p.ref))
            continue

        except (RowGeneratorError, MetatabFileNotFound):
            # directory is not a package, carry on
            pass

        for f in files:
            if not islink(join(root, f)):
                try:

                    p = open_package(join(root, f))
                    if str(p.ref) not in seen:
                        yield p
                    seen.add(str(p.ref))
                except (RowGeneratorError, MetatabFileNotFound):
                    # directory is not a package, carry on
                    pass


def write_s3(m):
    raise NotImplementedError()

    path = None

    print('Indexing tmp index', path)

    idx = SearchIndex(path)

    index_s3(m.s3_url, idx)

    with open(path) as f:
        m.bucket.write(f.read(), 'index.json', m.acl)


def index(args):
    n_indexed = 0
    write_index = False

    if args.write and args.list:
        err("Can't combine --write and --list")

    if args.write:
        if not args.file:
            args.file = join(tempfile.mkdtemp(), 'something')
            args.file = '/tmp/index.json'

    idx = SearchIndex(args.file or search_index_file())

    u = parse_app_url(args.metatab_url)

    if not isinstance(u, S3Url) and args.write:
        err("--write can only be used with S3 urls")

    if args.list:

        pkg_list = []
        for p in walk_packages(args, u):
            pkg_list.append((p.name, p.ref))

        prt(tabulate(pkg_list, headers='name Url'.split()))
    elif args.config:
        prt(idx.path)

    elif args.clear:
        idx.clear()
        prt('Cleared the index')
    elif args.dump:
        dump_index(args, idx)
    elif args.load:
        load_index(args, idx)
    elif isinstance(u, FileUrl):
        entries = []
        for p in walk_packages(args, u):
            if p.ref.get_resource().get_target().target_format == 'ipynb':
                continue
            prt('Adding: ', p.ref)
            idx.add_package(p)
            entries.append(p.name)
            n_indexed += 1

        write_index = True

        prt("Indexed ", len(entries), 'entries')

    elif isinstance(u, S3Url):
        index_s3(u, idx)

        if args.write:
            from metapack.package.s3 import S3Bucket

            acl = 'public-read'

            bucket = S3Bucket(u, acl=acl, profile=args.profile)

            with open(args.file, mode='rb') as f:
                bucket.write(f.read(), 'index.json', acl=acl)

    else:
        err(f"Can only index File and S3 urls, not {type(u)}")

    if args.result and n_indexed > 0:
        print(f"🗂 Indexed {n_indexed} packages")

    if write_index:
        idx.write()


def index_s3(u, idx):
    # S3 package collections are flat, so we don't have to walk recursively.
    # However, we are only going to take the S3 packages, because they have the distribution
    # information for the rest of the packages, so we can get info about Excel and Zip packages
    # without opening them.
    entries = []
    import re

    for e in u.list():

        if e.target_format == 'csv':

            prt("Processing ", e)

            p = open_package(e)

            for d in p.find('Root.Distribution'):
                u = parse_app_url(d.value)

                version_m = re.search('-([^-]+)$', p.name)

                if u.target_format in ('xlsx', 'zip', 'csv'):
                    idx.add_entry(p.identifier, p.name, p.nonver_name, version_m.group(1), u.target_format,
                                  'metapack+' + str(u))

                    entries.append(p.name)
    idx.write()

    prt("Indexed ", len(entries), 'entries to ', idx.path)


def dump_index(args, idx):
    """Create a metatab file for the index"""

    from metatab import MetatabDoc

    doc = MetatabDoc()

    pack_section = doc.new_section('Packages', ['Identifier', 'Name', 'Nvname', 'Version', 'Format'])

    r = doc['Root']
    r.new_term('Root.Title', 'Package Index')

    for p in idx.list():
        pack_section.new_term('Package',
                              p['url'],
                              identifier=p['ident'],
                              name=p['name'],
                              nvname=p['nvname'],
                              version=p['version'],
                              format=p['format'])

    doc.write_csv(args.dump)


def load_index(args, idx):
    from metatab import MetatabDoc

    doc = MetatabDoc(args.load)

    entries = set()

    for t in doc['Packages']:
        idx.add_entry(t.identifier, t.name, t.nvname, t.version, t.format, t.value)
        entries.add(t.value)

    prt("Loaded {} packages".format(len(entries)))

    idx.write()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

"""

import json
from os import rename
from os.path import exists
from shutil import copy


def search_index_file():
    """Return the default local index file, from the download cache"""
    from metapack import Downloader
    from os import environ

    return environ.get('METAPACK_SEARCH_INDEX',
                       Downloader.get_instance().cache.getsyspath('index.json'))


class SearchIndex(object):
    pkg_format_priority = {
        'fs': 6,
        'zip': 5,
        'xlsx': 4,
        'csv': 3,
        'web': 2,
        'source': 1,
        'unk': 0
    }

    def __init__(self, path):
        self.path = path

        self._db = None

    def open(self):
        if not self._db:
            try:
                with open(self.path) as f:
                    self._db = json.load(f)
            except FileNotFoundError:
                self._db = {}

        assert self._db is not None

    def clear(self):

        self._db = {}
        self.write()

    def write(self):
        """Safely write the index data to the index file """
        index_file = self.path
        new_index_file = index_file + '.new'
        bak_index_file = index_file + '.bak'

        if not self._db:
            return

        with open(new_index_file, 'w') as f:

            json.dump(self._db, f, indent=4)

        if exists(index_file):
            copy(index_file, bak_index_file)

        rename(new_index_file, index_file)

    def _make_package_entry(self, ident, name, nvname, version, format, url):
        """

        """

        self.open()

        try:
            version = 'V' + str(version).zfill(12)
        except (ValueError, TypeError):
            # No version, skip it.
            return

        if format is None:
            format = 'unk'
            return

        self._db[ident] = {'t': 'ident', 'ref': nvname}  # these should always be equivalent

        self._db[name] = {'t': 'name', 'ref': nvname, 'version': version, 'ident': ident}

        if nvname not in self._db:
            self._db[nvname] = {
                't': 'nvname',
                'packages': {}
            }

        key = '{}-{}'.format(name, format)

        self._db[nvname]['packages'][key] = {
            'name': name,
            'nvname': nvname,
            'version': version,
            'format': format,
            'ident': ident,
            'url': url
        }

    def add_package(self, pkg, format=None):
        from os.path import abspath

        ref_url = pkg.package_url.clone()
        ref_url.path = abspath(ref_url.path)

        ref = str(ref_url)

        nv_name = pkg._generate_identity_name(mod_version=None)  # Non versioned-name, for the latest package
        name = pkg.name
        version = pkg.get_value('Root.Version')
        identifier = pkg.get_value('Root.Identifier')

        target_ref = ref_url.get_resource().get_target()

        if format is None:
            if target_ref.fspath.is_dir() and target_ref.fspath.joinpath('metadata.csv').exists():
                if not pkg.get_value('Root.Issued') or target_ref.fspath.joinpath('_packages').exists():
                    format = 'source'
                else:
                    format = 'fs'
            else:
                format = target_ref.target_format

        self._make_package_entry(identifier, name, nv_name, version, format, ref)

    def add_entry(self, ident, name, nvname, version, format, url):
        self._make_package_entry(ident, name, nvname, version, format, url)

    def update(self, o):
        """Update from another index or index dict"""

        self.open()

        try:
            self._db.update(o._db)
        except AttributeError:
            self._db.update(o)

    def list(self):

        packages = []

        self.open()

        for k, v in self._db.items():
            if v.get('t') == 'nvname':
                for pkey, e in v.get('packages').items():
                    packages.append(e)

        return list(
            reversed(sorted(packages, key=lambda x: (x['name'], x['version'], self.pkg_format_priority[x['format']]))))

    def records(self):
        self.open()
        for k, v in self._db.items():
            if v['t'] == 'nvname':
                for pk, pv in v['packages'].items():
                    yield [pv['name'], pv['nvname'], pv['version'], pv['format'], pv['url']]

    def search(self, key, format='issued'):
        from rowgenerators import parse_app_url

        url = parse_app_url(key)

        search_term = url.path

        self.open()

        if format == 'all':
            format = None
        elif format == 'issued':  # 'issued' means  'not source'
            format = ['zip', 'csv', 'xlsx', 'fs', 'web']

        if format and not isinstance(format, (list, tuple)):
            format = [format]

        record = self._db.get(search_term)  # Case when the term is a name, nvname, or ident

        if record:
            records = [record]
            match_type = 'exact'
        else:
            records = [record for k, record in self._db.items() if search_term in k]
            match_type = 'subset'

        packages = []
        seen = set()
        for e in records:

            key_type = e['t']

            if key_type == 'ident':
                match_key = 'nvname'
                match_value = e['ref']
            else:
                match_key = key_type
                match_value = search_term

            if e.get('ref'):
                e = self._db[e['ref']]

            for pkey, p in e['packages'].items():

                if (match_type == 'exact' and p[match_key] == match_value) or \
                        (match_type == 'subset' and search_term in p[match_key]):
                    if (format and p['format'] in format) or format is None:
                        if not (p['format'], p['name']) in seen:
                            packages.append(p)
                            seen.add((p['format'], p['name']))

        return list(reversed(sorted(packages, key=lambda x: (x['version'], self.pkg_format_priority[x['format']]))))



# Copyright (c) 2019 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""Exported objects from this module are added to the document environment,
which is passed into generators. So, these functions can be used in 'python:'
scheme urls. """


def colmap(resource, doc, env, colmap, *args, **kwargs):
    """
    Run resources with column maps and combine multiple resources

    This function is referenced from a resource or reference url:

    Reference: python:#colmaps&colmap=data_group

    When run, the function will collect all of the other resources or references
    that have a ColMap arg prop of 'data_group' and iterate them after applying
    the column map.

    The function can take two additional arguments: ``prefix`` and ``suffix``,
    which name an arg property of the collected resources that will be
    prepended or appended to the yielded rows. For instance:

    Reference: data/data-1.csv
    Reference.Name: data_ref_1
    Reference.ColMap: data_group
    Reference.Year: 1999
    Reference.Month: 12

    Reference: python:#colmap&colmap=data_group&prefix=Year
    Reference.Name: collected
    Reference.Suffix: Month

    Running the ``collected`` reference will iterate over ``data_ref_1``,
    apply the column map in the file named ``colmap-data_group.csv``, alter the
    iterated header by prepending 'Year' and apppending 'Month', and alter
    the iterated data rows by prepending '1999' and appending '12'


    :param resource:
    :param doc:
    :param env:
    :param colmap:
    :param args:
    :param kwargs:
    :return:
    """

    from itertools import chain
    from rowgenerators.source import ReorderRowGenerator

    headers = None

    prefix = kwargs.get('prefix', resource.get_value('prefix', ''))
    pre_header = prefix.split(',')

    suffix = kwargs.get('suffix', resource.get_value('suffix', ''))
    post_header = suffix.split(',')

    for r in chain(doc.references(), doc.resources()):
        if r.get_value('ColMap') == colmap:

            pre_row = tuple([r.get_value(e) for e in pre_header])
            post_row = tuple([r.get_value(e) for e in post_header])

            rrg = ReorderRowGenerator(r, r.header_map)
            itr = iter(rrg)

            if not headers:
                headers = next(itr)
                yield pre_header + headers + post_header
            else:
                next(itr)

            for row in itr:
                yield pre_row + row + post_row


__all__ = [colmap]



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

import argparse
import json
import sys
from textwrap import dedent

from rowgenerators import parse_app_url
from rowgenerators.exceptions import AppUrlError

from metapack import Downloader
from metapack.cli.core import err, prt
from metapack.index import SearchIndex, search_index_file

downloader = Downloader.get_instance()

def search(subparsers):
    """Index packages for searching.

    The index file is a JSON file, which is by default index.json in the cache.
    The file can be moved by setting the METAPACK_SEARCH_INDEX environmental variable.

    """
    parser = subparsers.add_parser(
        'search',
        description=search.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-l', '--list', default=False, action='store_true',
                        help="List the packages in the index")

    parser.add_argument('-f', '--format', help='Select a specific package format')

    parser.add_argument('-1', '--one', action='store_true',
                        help='Find only one result, using the same resolution process used when building packages')

    parser.add_argument('-p', '--path', default=False, action='store_true',
                        help="Instead of displaying the Metapack URL, display only the local file path")

    parser.add_argument('-c', '--config', default=False, action='store_true',
                        help="Show the path to the index file")

    parser.add_argument('-j', '--json', default=False, action='store_true',
                       help="Output json for some commands")

    parser.add_argument('-n', '--name', default=False, action='store_true',
                       help="Output only the first column, the name of the package")

    parser.add_argument('-r', '--ref', default=False, action='store_true',
                       help="Output only the third column, the url of the package")

    parser.set_defaults(run_command=run_search)

    parser.add_argument('search', nargs='?', help="Path or URL to a metatab file")


def maybe_path(args,e):
    if args.path:
        u = parse_app_url(e).inner
        if u.proto == 'file':
            return u.fspath
        else:
            return ''
    else:
        return e

def run_search(args):

    from tabulate import tabulate

    if args.config:
        prt(search_index_file())
        sys.exit(0)

    if not args.search or args.list:

        idx = SearchIndex(search_index_file())

        if args.json:

            packages = [e for e in idx.list() if args.format is None or args.format == e['format']]

            print(json.dumps(packages))
        else:

            packages = [(e['name'], e['format'], maybe_path(args, e['url'])) for e in idx.list()
                        if args.format is None or args.format == e['format']]

            if args.name:
                for name, _, _ in packages:
                    print(name)
            elif args.ref:
                for _, _, url in packages:
                    print(url)
            else:
                print(tabulate(packages, headers='Name Format Url'.split()))

    elif args.one:

        if args.search.startswith('index'):
            url_s = args.search
        else:
            url_s = 'index:' + args.search

        try:
            u = parse_app_url(url_s)

            prt(str(maybe_path(args,u.get_resource())))

        except AppUrlError as e:
            err(f"Failed to resolve: {str(u)}; {str(e)}")

    else:

        idx = SearchIndex(search_index_file())


        p = idx.search(args.search, args.format)

        packages = []
        for e in p:
            packages.append((e['name'], e['format'], e['url']))

        if args.name:
            for name, _, _ in packages:
                print(name)
        elif args.ref:
            for _, _, url in packages:
                print(url)
        else:
            print(tabulate(packages, headers='Name Format Url'.split()))



import unittest
from textwrap import dedent

from metatab.rowgenerators import TextRowGenerator

from metapack import Downloader, MetapackDoc

downloader = Downloader()


class TestIssues(unittest.TestCase):
    """Test Metapack AppUrls and Row Generators"""

    def setUp(self):
        import warnings
        warnings.simplefilter('ignore')

    def test_integer_version(self):
        mt_lines = dedent("""
        Declare: metatab-latest
        Identifier: foobar

        Dataset: foobar
        Origin: foo.com
        Version: 1


        Section: References

        """)

        doc = MetapackDoc(TextRowGenerator(mt_lines))

        self.assertEqual('foo.com-foobar-1', doc._generate_identity_name())
        self.assertEqual('foo.com-foobar-1', doc._generate_identity_name(False))
        self.assertEqual('foo.com-foobar', doc._generate_identity_name(None))
        self.assertEqual('foo.com-foobar-10', doc._generate_identity_name(10))

        self.assertEqual('foo.com-foobar', doc.nonver_name)

        # Semantic

        mt_lines = dedent("""
        Declare: metatab-latest
        Identifier: foobar

        Dataset: foobar
        Origin: foo.com
        Version:
        Version.Major: 1
        Version.Minor: 2
        Version.Patch: 3

        Section: References

        """)

        doc = MetapackDoc(TextRowGenerator(mt_lines))

        self.assertEqual('foo.com-foobar-1.2.3', doc._generate_identity_name())
        self.assertEqual('foo.com-foobar-1.2.3', doc._generate_identity_name(False))
        self.assertEqual('foo.com-foobar', doc._generate_identity_name(None))
        self.assertEqual('foo.com-foobar-1.2.10', doc._generate_identity_name(10))

        self.assertEqual('foo.com-foobar', doc.nonver_name)

        mt_lines = dedent("""
        Declare: metatab-latest
        Identifier: foobar

        Dataset: foobar
        Origin: foo.com
        Version:
        Version.Major: 1

        Section: References

        """)

        doc = MetapackDoc(TextRowGenerator(mt_lines))

        self.assertEqual('foo.com-foobar-1.1.1', doc._generate_identity_name())
        self.assertEqual('foo.com-foobar-1.1.1', doc._generate_identity_name(False))
        self.assertEqual('foo.com-foobar', doc._generate_identity_name(None))
        self.assertEqual('foo.com-foobar-1.1.10', doc._generate_identity_name(10))

        self.assertEqual('foo.com-foobar', doc.nonver_name)


if __name__ == '__main__':
    unittest.main()
